import click
from mdkits.util import arg_type, os_operation


def common_setting(f):
    f = click.argument('filename', type=click.Path(exists=True), default=os_operation.default_file_name('*-pos-1.xyz', last=True))(f)
    f = click.option('-r', type=arg_type.FrameRange, help='range of frame to analysis')(f)
    f = click.option('--angle', type=(float, float), help='update water angle judgment')(f)
    f = click.option('--distance', type=float, help='update water distance judgment', default=1.2, show_default=True)(f)
    f = click.option('--update_water', is_flag=True, help='update water with distance or angle judgment')(f)
    f = click.option("--surface", type=str, help="surface group")(f)
    f = click.option('--cell', type=arg_type.Cell, help="set cell, a list of lattice, --cell x,y,z or x,y,z,a,b,c")(f)

    return f


#!/usr/bin/env python3

import click, os
from ase import build
import numpy as np
from mdkits.util import out_err


def surface_check(obj, surface_type):
    if hasattr(obj, surface_type):
        return getattr(obj, surface_type)


@click.command(name='surface')
@click.argument('symbol', type=str)
@click.argument('surface', type=click.Choice(['fcc100', 'fcc110', 'fcc111', 'fcc211', 'bcc100', 'bcc110', 'bcc111', 'hcp0001', 'hcp10m10', 'diamond100', 'diamond111', 'mx2', 'graphene']))
@click.argument('size', type=click.Tuple([int, int, int]))
@click.option('--kind', type=click.Choice(['2H', '1T']), help='designate the kind of MX2 surface')
@click.option('-a', type=float, help='the lattice constant. if specified, it overrides the expermental lattice constant of the element. must be specified if setting up a crystal structure different from the one found in nature')
@click.option('-c', type=float, help='extra hcp lattice constant. if specified, it overrides the expermental lattice constant of the element. Default is ideal ratio: sqrt(8/3)', default=np.sqrt(8/3), show_default=True)
@click.option('--thickness', type=float, help='Thickness of the layer, for mx2 and graphene')
@click.option('--orth', is_flag=True, help='if specified and true, forces the creation of a unit cell with orthogonal basis vectors. if the default is such a unit cell, this argument is not supported')
@click.option('--vacuum', type=float, help='designate vacuum of surface, default is None', default=0.1, show_default=True)
def main(symbol, surface, size, kind, a, c, thickness, orth, vacuum):
    """build a common surface"""

    vacuum = vacuum / 2
    build_surface = surface_check(build, surface)
    out_filename = f"{symbol}_{surface}_{size[0]}{size[1]}{size[2]}.cif"
    if surface in ['fcc100']:
        orth = True

    if surface in ['hcp0001', 'hcp10m10']:
        atoms = build_surface(symbol, size, a=a, c=c, vacuum=vacuum, orthogonal=orth)
    elif surface in ['mx2']:
        if thickness is None:
            atoms = build_surface(symbol, size, kind=kind, a=a, vacuum=vacuum)
        else:
            atoms = build_surface(symbol, size, kind=kind, a=a, vacuum=vacuum, thickness=thickness)
    elif surface in ['graphene']:
        if a is None:
            a = 2.46

        if thickness is None:
            atoms = build_surface(formula=symbol, size=size, a=a, vacuum=vacuum)
        else:
            atoms = build_surface(formula=symbol, size=size, thickness=thickness, a=a, vacuum=vacuum)
    else:
        atoms = build_surface(symbol, size, a=a, vacuum=vacuum, orthogonal=orth)
    
    out_err.check_cell(atoms)
    atoms.write(out_filename)
    print(os.path.abspath(out_filename))


if __name__ == '__main__':
    main()




import click
from mdkits.md_cli import (
    wrap,
    dipole,
    angle,
    density,
    hb_distribution,
    vac,
    rdf,
    msd,
    monitor,
)


@click.group(name='md')
@click.pass_context
def cli(ctx):
    """kits for MD analysis"""

cli.add_command(wrap.main)
cli.add_command(density.main)
cli.add_command(dipole.main)
cli.add_command(angle.main)
cli.add_command(hb_distribution.main)
cli.add_command(vac.main)
cli.add_command(rdf.main)
cli.add_command(msd.main)
cli.add_command(monitor.main)


if __name__ == '__main__':
    cli()


import os, shutil, atexit


def make_temp_dir(temp_dir, delete=False):
    if os.path.exists(temp_dir):
        shutil.rmtree(temp_dir)
    os.mkdir(temp_dir)
    if delete:
        atexit.register(remove_temp_dir, temp_dir)


def remove_temp_dir(temp_dir):
    shutil.rmtree(temp_dir)


def default_file_name(match, last=False):
    import glob
    file_list = glob.glob(match)
    if file_list:
        if last:
            default_file_name = sort_word_and_number(file_list)[-1]
        else:
            default_file_name = list(file_list)
    else:
        default_file_name = None

    return default_file_name


def sort_word_and_number(unsort_list):
    import re
    fns = lambda s: sum(((s,int(n))for s,n in re.findall('(\D+)(\d+)','a%s0'%s)),())
    sorted_list = sorted(unsort_list, key=fns)

    return sorted_list


def default_input():
    default_input_name = os.environ.get("DEFAULT_INPUT", "input.inp,setup.inp,cell.inc").split(',')
    return default_input_name


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np
from matplotlib import pyplot as plt
from matplotlib import rcParams
import matplotlib.ticker as ticker
import yaml, click
from mdkits.util import fig_operation

DEFAULT_CONFIG = {
        'data': {},
        'x': {},
        'y': {},
        'figsize': None,
        'x_range': None,
        'y_range': None,
        'legend_fontsize': None,
        'ticks_fontsize': None,
        'line': {},
        'fold': None
    }


def load_config(file_path):
    try:
        with open(file_path, 'r') as file:
            data = yaml.safe_load(file)
    except FileNotFoundError:
        print(f"no file {file_path}")
        exit(1)
    except yaml.YAMLError as exc:
        print(f"yaml parse error: {exc}")
        exit(1)
    config = data.copy()
    #for key, default in DEFAULT_CONFIG.items():
    #    config[key] = data.get(key, default)
    for key, value in config.items():
        for dkey, dvalue in DEFAULT_CONFIG.items():
            value[dkey] = data[key].get(dkey, dvalue)

    return config


def check_figsize(figsize):
    match figsize:
        case list() if len(figsize) == 2:
            rcParams['figure.figsize'] = figsize
        case None:
            pass
        case _:
            print("wrong yaml structure")
            exit(1)


def check_int_ticks(ax, int_ticks):
    match int_ticks:
        case 'x':
            ax.xaxis.set_major_locator(ticker.MaxNLocator(integer=True))
        case 'y':
            ax.yaxis.set_major_locator(ticker.MaxNLocator(integer=True))
        case 'a':
            ax.xaxis.set_major_locator(ticker.MaxNLocator(integer=True))
            ax.yaxis.set_major_locator(ticker.MaxNLocator(integer=True))
        case _:
            pass


def check_axis_range(ax, x_range=None, y_range=None):
    if x_range:
        match x_range:
            case list() if len(x_range) == 2:
                ax.set_xlim(x_range[0], x_range[1])
            case None:
                pass
            case _:
                print("wrong yaml structure")
                exit(1)
    if y_range:
        match y_range:
            case list() if len(y_range) == 2:
                ax.set_ylim(y_range[0], y_range[1])
            case None:
                pass
            case _:
                print("wrong yaml structure")
                exit(1)


def check_lenend_fontsize(ax, legend_fontsize, rmse=None):
    if rmse is not None:
        match legend_fontsize:
            case int():
                l = ax.legend(fontsize=legend_fontsize, markerscale=0, handlelength=0)
            case None:
                l = ax.legend(markerscale=0, handlelength=0)
            case _:
                pass

        #p = l.get_window_extent()
        #ax.annotate(f"RMSE:{rmse:.4f}", (p.p0[0], p.p1[1]), (p.p0[0], p.p1[1]-300), xycoords='figure pixels', fontsize=legend_fontsize)
    else:
        match legend_fontsize:
            case int():
                ax.legend(fontsize=legend_fontsize)
            case None:
                ax.legend()
            case _:
                pass


def check_output_mode(fig, plt, m, figure, fold=None):
    match m:
        case 'show':
            plt.show()
        case 'save':
            fig_operation.savefig(fig, f"{figure}", fold)
            plt.close()
        case 'all':
            fig_operation.savefig(fig, f"{figure}", fold)
            plt.show()
            plt.close()
        case _:
            pass


def check_ticks_fontsize(ax, ticks_fontsize):
    match ticks_fontsize:
        case int():
            ax.tick_params(axis='both', labelsize=ticks_fontsize)
        case None:
            pass
        case _:
            pass


def mode_1(data_dict, figure, **kwargs):
    # init config
    check_figsize(data_dict['figsize'])

    fig, ax = plt.subplots()

    # ticks config
    check_int_ticks(ax, kwargs['int_ticks'])

    # plot part
    for label, data in data_dict['data'].items():
        label = r"$\mathrm{ "+ label +" }$"
        data = np.loadtxt(data)
        x_slice, xlabel = next(iter(data_dict['x'].items()))
        for y_slice, ylabel in data_dict['y'].items():
            ax.plot(data[:, x_slice], data[:, y_slice], label=label)

        # range config
        check_axis_range(ax, x_range=data_dict['x_range'], y_range=data_dict['y_range'])

        # axis label part
        ax.set_xlabel(r"$\mathrm{ "+ xlabel +" }$")
        ax.set_ylabel(r"$\mathrm{ "+ ylabel +" }$")

        # h or v line
        match data_dict['line']:
            case _:
                pass

        # legend config
        check_lenend_fontsize(ax, data_dict['legend_fontsize'])

    # output part
    check_output_mode(fig, plt, kwargs['m'], figure, data_dict['fold'])


def mode_2(data_dict, figure, **kwargs):
    # init config
    check_figsize(data_dict['figsize'])

    fig, ax = plt.subplots()

    # ticks config
    check_int_ticks(ax, kwargs['int_ticks'])

    # plot part
    ylabel, data = next(iter(data_dict['data'].items()))
    data = np.loadtxt(data)
    x_slice, xlabel = next(iter(data_dict['x'].items()))
    for y_slice, label in data_dict['y'].items():
        label = r"$\mathrm{ "+ label +" }$"
        ax.plot(data[:, x_slice], data[:, y_slice], label=label)

        # range config
        check_axis_range(ax, x_range=data_dict['x_range'], y_range=data_dict['y_range'])

        # axis label part
        ax.set_xlabel(r"$\mathrm{ "+ xlabel +" }$")
        ax.set_ylabel(r"$\mathrm{ "+ ylabel +" }$")

    # legend config
    check_lenend_fontsize(ax, data_dict['legend_fontsize'])

    # output part
    check_output_mode(fig, plt, kwargs['m'], figure, data_dict['fold'])


def mode_error(data_dict, figure, **kwargs):
    # init config
    check_figsize(data_dict['figsize'])

    fig, ax = plt.subplots(figsize=(4.8, 4.8))

    # ticks config
    check_int_ticks(ax, kwargs['int_ticks'])

    # plot part
    label, data = next(iter(data_dict['data'].items()))
    data = np.loadtxt(data)
    x_slice, xlabel = next(iter(data_dict['x'].items()))
    for y_slice, ylabel in data_dict['y'].items():
        label = r"$\mathrm{ "+ label +" }$"
        ax.xaxis.set_major_formatter(ticker.FormatStrFormatter('%.2f'))
        ax.yaxis.set_major_formatter(ticker.FormatStrFormatter('%.2f'))
        ax.scatter(data[:, x_slice], data[:, y_slice], label=label)
        data_diff = data[:, x_slice] - data[:, y_slice]
        data_rmse = np.sqrt(np.average(data_diff*data_diff))
        ax.axline((data[0, x_slice], data[0, x_slice]), slope=1, ls='--', c='black')
        ax.set_xticks([])
        ax.set_yticks([])
        check_lenend_fontsize(ax, data_dict['ticks_fontsize'])

        # range config
        check_axis_range(ax, x_range=data_dict['x_range'], y_range=data_dict['y_range'])

        # axis label part
        ax.set_xlabel(r"$\mathrm{ "+ xlabel +" }$")
        ax.set_ylabel(r"$\mathrm{ "+ ylabel +" }$")

        lims = [np.min([ax.get_xlim(), ax.get_ylim()]), np.max([ax.get_xlim(), ax.get_ylim()])]
        ax.text(lims[0]+(lims[1]-lims[0])*.6, lims[0]+(lims[1]-lims[0])*.1, f"RMSE: {data_rmse:.4f}", fontsize=data_dict['legend_fontsize'])
        ax.set_xlim(lims)
        ax.set_ylim(lims)
        #ax.set_aspect('equal')
    #plt.axis('equal')
    # legend config
    check_lenend_fontsize(ax, data_dict['legend_fontsize'], rmse=data_rmse)

    # output part
    check_output_mode(fig, plt, kwargs['m'], figure, data_dict['fold'])


@click.command(name='plot')
@click.argument('yaml_file', type=click.Path(exists=True), nargs=-1)
@click.option('--int_ticks', help='set x ticks or y ticks with int, choices: x, y, a(all)', type=click.Choice(['x', 'y', 'a']))
@click.option('--error', help='set error mode', is_flag=True)
@click.option('-m', help='output mode: show, save, all, default is save', type=click.Choice(['show', 'save', 'all']), default='save')
def main(yaml_file, int_ticks, error, m):
    """
    a plot tool, read yaml file to plot
    """

    kwargs = locals()
    figure_dict = {}
    for yaml_file in yaml_file:
        yaml_dict = load_config(yaml_file)
        figure_dict.update(yaml_dict)

        for figure, data_dict in figure_dict.items():
            check_data = len(data_dict['data'])

            # mode error
            if error:
                mode = 'error'
                mode_error(data_dict, figure, **kwargs)

            # mode 1
            elif check_data > 1:
                if len(data_dict['y']) > 1:
                    print("wrong yaml structure")
                    exit(1)
                mode = 1
                mode_1(data_dict, figure, **kwargs)

            # mode 2
            elif check_data == 1:
                mode = 2
                mode_2(data_dict, figure, **kwargs)

            else:
                print("wrong yaml structure")
                exit(1)
            print(f"{figure} is done with mode {mode}")


if __name__ == '__main__':
    main()



"""
filename: cp2k_input_parsing.py
function: prase cp2k file
"""


import sys
from mdkits.util import os_operation, out_err


def parse_cell():
    """
    function: parse cell information from cp2k input file
    parameter:
        cp2k_input_file: filename of cp2k input
    return:
        cell: list with 6 number
    """
    for file in os_operation.default_input():
        try:
            with open(file, 'r') as f:
                cell = []
                for line in f:
                    if "ABC" in line:
                        xyz = line.split()[-3:]
                        cell.extend(xyz)
                    if "ALPHA_BETA_GAMMA" in line:
                        abc = line.split()[-3:]
                        cell.extend(abc)
                if len(cell) == 3:
                    cell.extend([90.0, 90.0, 90.0])

                out_err.cell_output(cell)
                return cell
        except FileNotFoundError:
            return [0., 0., 0., 90., 90., 90.]


#def get_cell(cp2k_input_file, cell=None):
#    if cell is None:
#        cell = parse_cell(cp2k_input_file)
#    else:
#        cell = cell
#        if len(cell) == 3:
#            cell.extend([90.0, 90.0, 90.0])
#
#    return cell



#!/usr/bin/env python3

from ase.io import write
import click
import os
from mdkits.util import out_err, arg_type


@click.command(name='convert')
@click.argument('atoms', type=arg_type.Structure)
@click.option('-c', help='covert to cif', is_flag=True)
@click.option('-x', help='covert to xyz', is_flag=True)
@click.option('-d', help='covert to lammps data file', is_flag=True)
@click.option('-v', help='covert to vasp', is_flag=True)
@click.option('--coord', help='coord format', is_flag=True)
@click.option('--cp2k', help='convert to cp2k format(coord + cell)', is_flag=True)
@click.option('--center', help='center atoms', is_flag=True)
@click.option('--cell', type=arg_type.Cell, help='set cell, a list of lattice: --cell x,y,z or x,y,z,a,b,c')
def main(atoms, c, x, d, v, coord, cp2k, center, cell):
    """
    convet structure file in some formats
    """
    out_err.check_cell(atoms, cell)
    o = atoms.filename.split('.')[-2]


    if center:
        atoms.center()

    if c:
        o += '.cif'
        write(o, atoms, format='cif')

    if x:
        o += '.xyz'
        write(o, atoms, format='extxyz')

    if d:
        o += '.data'
        write(o, atoms, format='lammps-data', atom_style='atomic')

    if v:
        o = 'POSCAR'
        write(o, atoms, format='vasp')


    if cp2k:
        o = 'coord.xyz'
        write(o, atoms, format='xyz')
        with open(o, 'r') as f:
            lines = f.readlines()
        with open(o, 'w') as f:
            f.writelines(lines[2:])
        with open('cell.inc', 'w') as f:
            cell = atoms.cell.cellpar()
            f.write(f"ABC [angstrom] {cell[0]} {cell[1]} {cell[2]}\n")
            f.write(f"ALPHA_BETA_GAMMA {cell[3]} {cell[4]} {cell[5]}\n")


    print(os.path.abspath(o))


if __name__ == '__main__':
    main()



import click
import MDAnalysis as mda
import MDAnalysis.analysis.msd as msd
import numpy as np


@click.command(name="msd")
@click.argument("filename", type=click.Path(exists=True))
@click.argument('type', type=click.Choice(['xyz', 'xy', 'yz', 'xz', 'x', 'y', 'z']))
@click.argument("group", type=str)
def main(filename, type, group):
    """analysis msd along the given axis"""
    u = mda.Universe(filename)
    MSD = msd.EinsteinMSD(u, select=group, msd_type=type, fft=True)
    MSD.run(verbose=True)

    data = np.arange(1, MSD.n_frames + 1).reshape(-1, 1)
    s = "_"
    name = f"{s.join(group.split(' '))}"
    header = ''
    for i in range(MSD.n_particles):
        data = np.concatenate((data, MSD.results.msds_by_particle[:, i].reshape(-1, 1)), axis=1)
        header += name + f"_{i}\t"

    np.savetxt(f"msd_{type}.dat", data, fmt="%.5f", delimiter="\t", header=f"frame\t{header}")



if __name__ == '__main__':
    main()


import click
from .setting import common_setting
import MDAnalysis
from MDAnalysis import Universe
from MDAnalysis.analysis.base import AnalysisBase
from mdkits.util import arg_type, numpy_geo
import numpy as np
import sys


class Monitor(AnalysisBase):
    def __init__(self, filename, cell, index, surface, dt=0.001):
        u = Universe(filename)
        u.trajectory.ts.dt = dt
        u.dimensions = cell

        self.u = u
        self.atomgroup = self.u.select_atoms("all")
        self.surface = surface

        if index is None:
            sys.exit("Please specify at least one atom to monitor")
        else:
            self.index = index
            self.group = " or ".join([ f"index {i}" for i in self.index ])

        if surface is not None:
            self.surface_group = self.atomgroup.select_atoms(f"{surface}")
            if self.surface_group.n_atoms == 0:
                sys.exit("Please specify the correct surface group")
        else:
            self.surface_group = False

        super(Monitor, self).__init__(self.atomgroup.universe.trajectory, verbose=True)

    def _prepare(self):
        self.height = []
        if len(self.index) == 2:
            self.distance = []
        elif len(self.index) == 3:
            self.distance = []
            self.angle = []
    
    def _single_frame(self):
        if self.surface_group:
            surface = numpy_geo.find_surface(self.surface_group.positions[:, 2])
        else:
            surface = [0]

        self.height.append(self.atomgroup.select_atoms(self.group).positions[:, 2] - surface[0])

        if len(self.index) == 2:
            vec = MDAnalysis.lib.distances.minimize_vectors(self.atomgroup[self.index[0]].position - self.atomgroup[self.index[1]].position, self.u.dimensions)

            self.distance.append(np.linalg.norm(vec))

        if len(self.index) == 3:
            vec1 = MDAnalysis.lib.distances.minimize_vectors(self.atomgroup[self.index[0]].position - self.atomgroup[self.index[1]].position, self.u.dimensions)
            vec2 = MDAnalysis.lib.distances.minimize_vectors(self.atomgroup[self.index[2]].position - self.atomgroup[self.index[1]].position, self.u.dimensions)

            self.distance.append(np.array([np.linalg.norm(vec1), np.linalg.norm(vec2)]))
            self.angle.append(numpy_geo.vector_vector_angle(vec1, vec2))

    def _conclude(self):
        frame_count = np.arange(self.u.trajectory.n_frames).reshape(-1, 1)

        self.height = np.vstack(self.height)

        np.savetxt("monitor.dat", np.hstack((frame_count, self.height)), fmt="%.5f", header=f"frame\t{'  '.join(self.atomgroup.select_atoms(self.group).names)}")

        if len(self.index) == 2:
            self.distance = np.vstack(self.distance)

            np.savetxt("monitor.dat", np.hstack((frame_count, self.height, self.distance)), fmt="%.5f", header=f"frame\t\t{'  '.join(self.atomgroup.select_atoms(self.group).names)}\t{self.atomgroup.names[self.index[0]]}-{self.atomgroup.names[self.index[1]]}")
        elif len(self.index) == 3:
            self.distance = np.vstack(self.distance)
            self.angle = np.vstack(self.angle)

            np.savetxt("monitor.dat", np.hstack((frame_count, self.height, self.distance, self.angle)), fmt="%.5f", header=f"frame\t\t{'  '.join(self.atomgroup.select_atoms(self.group).names)}\t{self.atomgroup.names[self.index[0]]}-{self.atomgroup.names[self.index[1]]}\t{self.atomgroup.names[self.index[2]]}-{self.atomgroup.names[self.index[1]]}\t{self.atomgroup.names[self.index[0]]}-{self.atomgroup.names[self.index[1]]}-{self.atomgroup.names[self.index[2]]}")


@click.command(name="monitor")
@click.argument("filename", type=click.Path(exists=True))
@click.option("--index", "-i", type=int, help="index of atom to monitor", multiple=True)
@click.option('--cell', type=arg_type.Cell, help="set cell, a list of lattice, --cell x,y,z or x,y,z,a,b,c")
@click.option("--surface", type=str, help="surface group")
@click.option('-r', type=arg_type.FrameRange, help='range of frame to analysis')
def main(filename, cell, index, surface, r):
    """
    monitor the property of between atoms
    """
    a = Monitor(filename, cell, index, surface)
    if r is not None:
        if len(r) == 2:
            a.run(start=r[0], stop=r[1])
        elif len(r) == 3:
            a.run(start=r[0], stop=r[1], step=r[2])
    else:
        a.run()
    
    

if __name__ == "__main__":
    main()


#!/usr/bin/env python3

import numpy as np
import click
import MDAnalysis
from MDAnalysis import Universe
from MDAnalysis.analysis.base import AnalysisBase
from mdkits.util import numpy_geo, encapsulated_mda
import warnings, sys
from .setting import common_setting
warnings.filterwarnings("ignore")


class Hb_distribution(AnalysisBase):
    def __init__(self, filename, cell, surface, update_water, distance_judg, angle_judg, hb_distance, hb_angle, bin_size=0.2, dt=0.001, index=None):
        u = Universe(filename)
        u.trajectory.ts.dt = dt
        u.dimensions = cell
        self.u = u
        self.atomgroup = u.select_atoms("all")
        self.hb_distance = hb_distance
        self.hb_angle = hb_angle
        self.bin_size = bin_size
        self.surface = surface
        self.update_water = update_water
        self.frame_count = 0
        np.set_printoptions(threshold=np.inf)

        if surface is not None:
            self.surface_group = self.atomgroup.select_atoms(f"{surface}")
            if self.surface_group.n_atoms == 0:
                sys.exit("Please specify the correct surface group")
        else:
            self.surface_group = False

        if self.update_water:
            self.distance_judg = distance_judg
            self.angle_judg = angle_judg
        
        if index is not None:
            self.index = index
            self.hb_d_index = 0
            self.hb_a_index = 0
        else:
            self.index = None

        super(Hb_distribution, self).__init__(self.atomgroup.universe.trajectory, verbose=True)

    def _prepare(self):
        bin_num = int(self.u.dimensions[2] / self.bin_size) + 2
        self.accepter = np.zeros(bin_num, dtype=np.float64)
        self.donor = np.zeros(bin_num, dtype=np.float64)
        self.od = np.zeros(bin_num, dtype=np.float64)
        if self.surface_group:
            self.surface_pos = np.zeros(2)

    def _append(self, hb_d, hb_a, o):
        bins_d = np.floor(hb_d / self.bin_size).astype(int) + 1
        bins_a = np.floor(hb_a / self.bin_size).astype(int) + 1
        bins_o = np.floor(o / self.bin_size).astype(int) + 1

        bins_d = bins_d[bins_d < len(self.donor)]
        bins_a = bins_a[bins_a < len(self.accepter)]
        bins_o = bins_o[bins_o < len(self.od)]

        np.add.at(self.donor, bins_d, 1)
        np.add.at(self.accepter, bins_a, 1)
        np.add.at(self.od, bins_o, 1)

    def _single_frame(self):
        if self.update_water:
            o = self.atomgroup.select_atoms("name O")
            h = self.atomgroup.select_atoms("name H")

            o_group, oh1, oh2 = encapsulated_mda.update_water(self, o, h, distance_judg=self.distance_judg, angle_judg=self.angle_judg, return_index=False)

            o_pair = MDAnalysis.lib.distances.capped_distance(o_group.positions, o_group.positions, min_cutoff=0, max_cutoff=self.hb_distance, box=self.u.dimensions, return_distances=False)

            o0 = o_group[o_pair[:, 0]]
            o1 = o_group[o_pair[:, 1]]

            o0h1 = oh1[o_pair[:, 0]]
            o0h2 = oh2[o_pair[:, 0]]
            o1h1 = oh1[o_pair[:, 1]]
            o1h2 = oh2[o_pair[:, 1]]
        else:
            o_group = self.atomgroup.select_atoms("name O")
            o_pair = MDAnalysis.lib.distances.capped_distance(o_group.positions, o_group.positions, min_cutoff=0, max_cutoff=self.hb_distance, box=self.u.dimensions, return_distances=False)

            o0 = o_group[o_pair[:, 0]]
            o1 = o_group[o_pair[:, 1]]

            o0h1 = self.atomgroup[o0.indices + 1]
            o0h2 = self.atomgroup[o0.indices + 2]
            o1h1 = self.atomgroup[o1.indices + 1]
            o1h2 = self.atomgroup[o1.indices + 2]

        angle_o0h1_o0_o1 = np.degrees(
            MDAnalysis.lib.distances.calc_angles(o0h1.positions, o0.positions, o1.positions, box=self.u.dimensions)
        )
        angle_o0h2_o0_o1 = np.degrees(
            MDAnalysis.lib.distances.calc_angles(o0h2.positions, o0.positions, o1.positions, box=self.u.dimensions)
        )
        angle_o1h1_o1_o0 = np.degrees(
            MDAnalysis.lib.distances.calc_angles(o1h1.positions, o1.positions, o0.positions, box=self.u.dimensions)
        )
        angle_o1h2_o1_o0 = np.degrees(
            MDAnalysis.lib.distances.calc_angles(o1h2.positions, o1.positions, o0.positions, box=self.u.dimensions)
        )

        condition_d = (angle_o0h1_o0_o1 < self.hb_angle) | (angle_o0h2_o0_o1 < self.hb_angle)
        condition_a = (angle_o1h1_o1_o0 < self.hb_angle) | (angle_o1h2_o1_o0 < self.hb_angle)

        if self.index is not None:
            self.hb_d_index += o0.positions[:, 2][condition_d & (o0.indices == self.index)].shape[0]
            self.hb_a_index += o0.positions[:, 2][condition_a & (o0.indices == self.index)].shape[0]
        else:
            hb_d = o0.positions[:, 2][condition_d]
            hb_a = o0.positions[:, 2][condition_a]

            self._append(hb_d, hb_a, o_group.positions[:, 2])

        if self.surface_group:
            surface = numpy_geo.find_surface(self.surface_group.positions[:, 2])
            self.surface_pos[0] += surface[0]
            if len(surface) > 1:
                self.surface_pos[1] += surface[1]

        self.frame_count += 1

    def _conclude(self):
        if self.frame_count > 0 and self.index is None:
            average_od = self.od / self.frame_count
            average_donor = np.nan_to_num((self.donor / self.frame_count) / average_od, nan=0)
            average_accepter = np.nan_to_num((self.accepter / self.frame_count) / average_od, nan=0)
            average_sum = average_donor + average_accepter

            bins_z = np.arange(len(self.donor)) * self.bin_size

            if self.surface:
                surface = self.surface_pos/self.frame_count
                lower_z = surface[0]
                if surface[1] ==0:
                    upper_z = np.inf
                mask = (bins_z >= lower_z) & (bins_z <= upper_z)
                filtered_bins_z = bins_z[mask] - lower_z
                filtered_average_accepter = average_accepter[mask]
                filtered_average_donor = average_donor[mask]
                filtered_average_sum = average_sum[mask]

                combined_data = np.column_stack((filtered_bins_z, filtered_average_accepter, filtered_average_donor, filtered_average_sum))
            else:
                combined_data = np.column_stack((bins_z, average_accepter, average_donor, average_sum))

            np.savetxt("hb_distribution.dat", combined_data, header="Z\tAccepter\tDonor\tAccepter+Donor", fmt='%.5f', delimiter='\t')

        if self.index is not None and self.frame_count > 0:
            self.hb_d_index /= self.frame_count
            self.hb_a_index /= self.frame_count
            output = f"# {self.index}\naccepter     : {self.hb_a_index}\ndonor        : {self.hb_d_index}\ntotal        : {self.hb_a_index + self.hb_d_index}"
            with open(f"hb_{self.index}.dat", "a") as f:
                f.write(output)
            print(output)


@click.command(name="hb", help="analysis hydrogen bond distribution along z-axis")
@common_setting
@click.option('--hb_param', type=click.Tuple([float, float]), help='parameter for hydrogen bond', default=(3.5, 35), show_default=True)
@click.option('--index', type=int, help='index of an atom')
def main(filename, hb_param, cell, surface, r, update_water, distance, angle, index):
    """analysis hydrogen bond distribution along z-axis"""
    hb_dist = Hb_distribution(filename, cell, surface, update_water=update_water, distance_judg=distance, angle_judg=angle, hb_distance=hb_param[0], hb_angle=hb_param[1], index=index)

    if r is not None:
        if len(r) == 2:
            hb_dist.run(start=r[0], stop=r[1])
        elif len(r) == 3:
            hb_dist.run(start=r[0], stop=r[1], step=r[2])
    else:
        hb_dist.run()


if __name__ == '__main__':
    main()